package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.FileConfiguration
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.ConfigurationRepository
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import java.time.LocalDateTime
import java.util.*
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class ConfigurationServiceTest {
    private val repository = mockk<ConfigurationRepository>()
    private val fileConfiguration = mockk<FileConfiguration>()
    private val service = spyk(ConfigurationServiceImpl(repository, mockk(), fileConfiguration, mockk(), mockk(), mockk()))

    @AfterEach
    fun afterEach() {
        clearAllMocks()
    }

    // getAll()

    @Test
    fun `getAll() gets the Configuration of each ConfigurationType`() {
        every { service.get(any<ConfigurationType>()) } answers { throw ConfigurationNotSetException(this.args[0] as ConfigurationType) }

        service.getAll()

        verify {
            service.getAll()
            ConfigurationType.values().forEach {
                service.get(it)
            }
        }
        confirmVerified(service)
    }

    @Test
    fun `getAll() only returns set configurations`() {
        val unsetConfigurationTypes = listOf(
            ConfigurationType.INSTANCE_NAME,
            ConfigurationType.INSTANCE_LOGO_PATH,
            ConfigurationType.INSTANCE_ICON_PATH
        )

        every { service.get(match<ConfigurationType> { it in unsetConfigurationTypes }) } answers {
            throw ConfigurationNotSetException(this.firstArg() as ConfigurationType)
        }
        every { service.get(match<ConfigurationType> { it !in unsetConfigurationTypes }) } answers {
            val type = firstArg<ConfigurationType>()
            configuration(type, type.key)
        }

        val found = service.getAll()

        assertFalse {
            found.any {
                it.type in unsetConfigurationTypes
            }
        }

        verify {
            service.getAll()
            ConfigurationType.values().forEach {
                service.get(it)
            }
        }
        confirmVerified(service)
    }

    @Test
    fun `getAll() only includes configurations matching the formatted formatted key list`() {
        val configurationTypes = listOf(
            ConfigurationType.INSTANCE_NAME,
            ConfigurationType.INSTANCE_LOGO_PATH,
            ConfigurationType.INSTANCE_ICON_PATH
        )
        val formattedKeyList = configurationTypes
            .map { it.key }
            .reduce { acc, s -> "$acc$CONFIGURATION_FORMATTED_LIST_DELIMITER$s" }

        every { service.get(any<String>()) } answers {
            val key = firstArg<String>()
            configuration(key.toConfigurationType(), key)
        }

        val found = service.getAll(formattedKeyList)

        assertTrue {
            found.all { it.type in configurationTypes }
        }

        verify {
            service.getAll(formattedKeyList)
            configurationTypes.forEach {
                service.get(it.key)
            }
        }
        confirmVerified(service)
    }

    // get()

    @Test
    fun `get(key) calls get() with the ConfigurationType matching the given key`() {
        val type = ConfigurationType.INSTANCE_ICON_PATH
        val key = type.key

        every { service.get(type) } answers {
            val type = firstArg<ConfigurationType>()
            configuration(type, type.key)
        }

        service.get(key)

        verify {
            service.get(key)
            service.get(type)
        }
        confirmVerified(service)
    }

    @Test
    fun `get(type) gets in the repository when the given ConfigurationType is not computed or a file property`() {
        val type = ConfigurationType.INSTANCE_ICON_PATH

        every { repository.findById(type.key) } returns Optional.of(
            ConfigurationEntity(type.key, type.key, LocalDateTime.now())
        )

        val configuration = service.get(type)

        assertTrue {
            configuration.key == type.key
        }

        verify {
            service.get(type)
            repository.findById(type.key)
        }
        confirmVerified(service, repository)
    }

    @Test
    fun `get(type) gets in the FileConfiguration when the gien ConfigurationType is a file property`() {
        val type = ConfigurationType.DATABASE_URL

        every { fileConfiguration.get(type) } returns configuration(type, type.key)

        val configuration = service.get(type)

        assertTrue {
            configuration.key == type.key
        }

        verify {
            service.get(type)
            fileConfiguration.get(type)
        }
        verify(exactly = 0) {
            repository.findById(type.key)
        }
        confirmVerified(service, fileConfiguration, repository)
    }

    @Test
    fun `get(type) computes computed properties`() {
        val type = ConfigurationType.JAVA_VERSION

        val configuration = service.get(type)

        assertTrue {
            configuration.key == type.key
        }

        verify {
            service.get(type)
        }
        verify(exactly = 0) {
            repository.findById(type.key)
            fileConfiguration.get(type)
        }
        confirmVerified(service, repository, fileConfiguration)
    }

    @Test
    fun `get(type) throws ConfigurationNotSetException when the given ConfigurationType has no set configuration`() {
        val type = ConfigurationType.INSTANCE_ICON_PATH

        every { repository.findById(type.key) } returns Optional.empty()

        with(assertThrows<ConfigurationNotSetException> { service.get(type) }) {
            assertEquals(type, this.type)
        }

        verify {
            service.get(type)
            repository.findById(type.key)
        }
    }

    @Test
    fun `set() set the configuration in the FileConfiguration when the given ConfigurationType is a file configuration`() {
        val type = ConfigurationType.DATABASE_URL
        val content = "url"

        every { fileConfiguration.set(type, content) } just runs

        service.set(type, content)

        verify {
            service.set(type, content)
            fileConfiguration.set(type, content)
        }
        confirmVerified(service, fileConfiguration)
    }

    @Test
    fun `set() set the configuration in the repository when the given ConfigurationType is not a computed configuration of a file configuration`() {
        val type = ConfigurationType.INSTANCE_ICON_PATH
        val content = "path"
        val configuration = configuration(type, content)
        val entity = configuration.toEntity()

        every { repository.save(entity) } returns entity

        service.set(type, content)

        verify {
            service.set(type, content)
            repository.save(entity)
        }
        confirmVerified(service, repository)
    }

    @Test
    fun `set() throws CannotSetComputedConfigurationException when the given ConfigurationType is a computed configuration`() {
        val type = ConfigurationType.JAVA_VERSION
        val content = "5"

        with(assertThrows<CannotSetComputedConfigurationException> { service.set(type, content) }) {
            assertEquals(type, this.type)
        }

        verify {
            service.set(type, content)
        }
        confirmVerified(service)
    }
}
